
const char *
e_get_xterm_title (void) {
    static char title[4096];
    static int tty = -1;
    static int is_xterm = -1; /* -1 => unknown; 0 => no; 1 => yes */
    char *env_term = NULL;

    if (is_xterm == 0) {
        return (NULL);
    }
    else if (is_xterm == 1 && tty != -1) {
        char *t;
        if ((t = xterm_read_title(tty, title, sizeof(title))))
            return (t);
        else {
            close(tty); tty = -1;
            return (NULL);
        }
    }
    else if (is_xterm == 1 
                || ((env_term = getenv("TERM")) 
                    && (!strcmp(env_term, "xterm")
                        || !strcmp(env_term, "rxvt")))) {
        /* Terminal specified in $TERM is xterm compatible. */
        is_xterm = 1;
        if ((tty = open("/dev/tty", O_RDWR | O_NOCTTY)) != -1) {
            char *t;
            if ((t = xterm_read_title(tty, title, sizeof(title))))
                return (t);
            else {
                close(tty); tty = -1;
                return (NULL);
            }
        }
        else {
            /* Cannot open "/dev/tty". */
            return (NULL);
        }
    }
    else {
        /* Terminal specified in $TERM is _not_ xterm compatible. */
        is_xterm = 0;
        return (NULL);
    }
}

static void 
xterm_sig_alrm_h (int sig) {
    /* Do nothing.
     * The only purpose of this signal is to interrupt the 
     * read system call in xterm_read_title(). */
}

static char *
xterm_read_title (int tty, char *title, size_t len) {
    if (tty == -1 || title == NULL) {
        return (NULL);
    }
    else {
        struct sigaction sa;
        int cnt;
        char *cmd = "\033[21t";
        int e;
        
        /* Make sure we will not be blocked infinitely. */
        sa.sa_handler = xterm_sig_alrm_h;
        sa.sa_flags = SA_INTERRUPT | SA_RESETHAND;
        sigaction(SIGALRM, &sa, NULL);
        alarm(5);

        /* Write the request for window title. */
        e = strlen(cmd); 
        while ((cnt = write(tty, cmd, e)) > 0) {
            e -= cnt;
            cmd += cnt;
        }
        
        if (cnt == -1) {
            return (NULL);
        }
        else {
            /* Read the response. */
            int i = 0;
            char ch;
            int seen_esc = 0;
            while (i < len && (cnt = read(tty, &ch, 1)) > 0) {
                if (seen_esc == 0) {
                    if (ch == '\033')
                        seen_esc = 1;
                    else
                        continue;
                }
                if (ch == '\x9c')  /* ST */
                    break;
                title[i++] = ch;
            }
            if (cnt == -1 || i < 4) {
                return (NULL);
            }
            else {
                title[i] = '\0';
                return (title + 3); /* OSC l */
            }
        }
    }
}

